@@ -25,12 +25,27 @@ import einx
 from einops import rearrange, repeat, reduce, einsum, pack
 from einops.layers.torch import Rearrange
 
-from transfusion_pytorch.tensor_typing import Float, Int, Bool
-
 from rotary_embedding_torch import RotaryEmbedding, apply_rotary_emb
 
+from tqdm import tqdm
+
 pad_sequence = partial(pad_sequence, batch_first = True)
 
+# tensor typing
+
+import jaxtyping
+
+class TorchTyping:
+    def __init__(self, abstract_dtype):
+        self.abstract_dtype = abstract_dtype
+
+    def __getitem__(self, shapes: str):
+        return self.abstract_dtype[Tensor, shapes]
+
+Float = TorchTyping(jaxtyping.Float)
+Int   = TorchTyping(jaxtyping.Int)
+Bool  = TorchTyping(jaxtyping.Bool)
+
 # maybe flex attention
 
 try:
@@ -40,6 +55,8 @@ except ImportError:
 
 # constants
 
+ModalitySample = list[Int['_'] | Float['_ _'] | tuple[int, Float['_ _']]]
+
 RawModalityPositions = list[list[tuple[int, int]]]
 
 class LossBreakdown(NamedTuple):
@@ -67,6 +84,15 @@ def divisible_by(num, den):
 def cast_tuple(t, length = 1):
     return t if isinstance(t, tuple) else ((t,) * length)
 
+def eval_decorator(fn):
+    def inner(self, *args, **kwargs):
+        was_training = self.training
+        self.eval()
+        out = fn(self, *args, **kwargs)
+        self.train(was_training)
+        return out
+    return inner
+
 # tensor helpers
 
 def l2norm(t):
@@ -121,6 +147,10 @@ def modality_positions_to_tensor(
 
     modalities: list[Tensor] = [tensor(modality, device = device) for modality in modalities]
     modalities = pad_sequence(modalities, padding_value = pad_value)
+
+    if modalities.ndim == 2:
+        modalities = modalities.reshape(*modalities.shape, 3)
+
     return modalities
 
 # sanitizing modalities tensor, making sure it is ordered
@@ -135,11 +165,9 @@ def order_modality_positions_by_seq_offset(
     offsets_to_sort = offsets.masked_fill(no_modality_mask, 1e10)
     _, sorted_indices = offsets_to_sort.sort(dim = -1)
 
-    # sort by ascending offset and do a final mask of both offset and length to 0
+    # sort by ascending offset
 
     modalities = einx.get_at('b [mi] ..., b mo -> b mo ...', modalities, sorted_indices)
-    modalities = einx.where('b m, b m ..., -> b m ...', ~no_modality_mask, modalities, 0.)
-
     return modalities, sorted_indices
 
 # deriving relative positions from modality positions
@@ -239,6 +267,20 @@ def naive_attn_mask(
 
     return is_causal | is_modality.any(dim = 1)
 
+# sampling related functions
+
+# min_p for text
+# https://arxiv.org/abs/2407.01082
+
+def min_p_filter(logits, min_p = 0.1):
+    probs = logits.softmax(dim = -1)
+    max_probs = probs.amax(dim = -1, keepdim = True)
+    limit = min_p * max_probs
+    return torch.where(probs < limit, float('-inf'), logits)
+
+
+from torchdiffeq import odeint
+
 # random fourier embedding
 
 class RandomFourierEmbed(Module):
@@ -391,6 +433,7 @@ class Attention(Module):
         attn_mask: Tensor | None = None,
         rotary_emb: Tensor | None = None,
         block_mask = None,
+        return_kv_cache = False
     ):
         assert not (exists(block_mask) and exists(attn_mask))
 
@@ -398,6 +441,11 @@ class Attention(Module):
 
         q, k, v = self.to_qkv(x)
 
+        # maybe kv cache
+
+        if return_kv_cache:
+            kv_cache = torch.stack((k, v))
+
         # rotary embeddings
 
         if exists(rotary_emb):
@@ -432,7 +480,12 @@ class Attention(Module):
 
         # combine heads and out
 
-        return self.to_out(out)
+        out = self.to_out(out)
+
+        if not return_kv_cache:
+            return out
+
+        return out, kv_cache
 
 class Transformer(Module):
     def __init__(
@@ -534,6 +587,11 @@ class Transfusion(Module):
         modality_token_transform: tuple[str | callable, ...] | None = None,
         ignore_index = -1,
         diffusion_loss_weight = 1.,
+        odeint_kwargs: dict = dict(
+            atol = 1e-5,
+            rtol = 1e-5,
+            method = 'midpoint'
+        ),
     ):
         super().__init__()
 
@@ -560,12 +618,20 @@ class Transfusion(Module):
 
         # modality start and end tokens - termed [som] [eom] in this repo
 
-        num_som_eom_tokens = self.num_modalities * 2
+        num_som_eom_tokens = (self.num_modalities + 1) * 2
         som_eom_tensor = torch.arange(num_som_eom_tokens) + num_text_tokens # shift to the very end
         som_eom_tensor = rearrange(som_eom_tensor, '(be m) -> be m', be = 2)
 
-        self.som_ids, self.eom_ids = som_eom_tensor.tolist()
-        print(self.som_ids, self.eom_ids)
+        text_start_end_tensor, modality_start_end_tensor = som_eom_tensor[:, 0], som_eom_tensor[:, 1:]
+
+        # modality start and end ids
+
+        self.som_ids, self.eom_ids = modality_start_end_tensor.tolist()
+
+        # entire "sentence" start and end id
+
+        self.sos_id, self.eos_id = text_start_end_tensor.tolist()
+
         # modality transforms
 
         modality_token_transform = cast_tuple(modality_token_transform, self.num_modalities)
@@ -595,23 +661,142 @@ class Transfusion(Module):
         self.ignore_index = ignore_index
         self.diffusion_loss_weight = diffusion_loss_weight
 
+        # diffusion sampling related
+
+        self.odeint_fn = partial(odeint, **odeint_kwargs)
+
     @property
     def device(self):
         return next(self.parameters()).device
 
+    @torch.no_grad()
+    @eval_decorator
+    def sample(
+        self,
+        prompt: ModalitySample | None = None,
+        max_length = 8192,
+        text_temperature = 1.5,
+        text_min_p = 0.1,
+        modality_length = 32, # fix the modality token length for now, but this will be determined by the language model in a metadata tag
+        modality_steps = 16
+    ) -> ModalitySample:
+
+        device = self.device
+
+        init_text_seq = tensor([self.sos_id], device = self.device)
+        modality_sample = [init_text_seq]
+
+        curr_length = 0
+        curr_modality_id = None
+        num_past_modalities = 0  # starts off with no modalities in output
+        is_decoding_text = True  # starts off with text decoding, and alternates with modalities depending on [som] tokens detected
+
+        with tqdm(total = max_length) as pbar:
+
+            while curr_length <= max_length:
+
+                if is_decoding_text:
+                    pbar.set_description('decoding text')
+
+                    *_, seq = modality_sample
+
+                    logits = self.forward([modality_sample], return_loss = False)
+                    logits = logits[0][-1]
+
+                    logits = min_p_filter(logits, min_p = text_min_p)
+                    probs = (logits / text_temperature).softmax(dim = -1)
+
+                    sampled = torch.multinomial(probs, 1)
+
+                    seq = torch.cat((seq, sampled), dim = -1)
+                    modality_sample[-1] = seq
+
+                    pbar.update(1)
+                    curr_length += 1
+
+                    sampled_token_id = sampled.item()
+
+                    if sampled_token_id == self.eos_id:
+                        break
+
+                    if sampled_token_id in self.som_ids:
+                        curr_modality_id = self.som_ids.index(sampled_token_id)
+                        is_decoding_text = False
+
+                else:
+                    assert exists(curr_modality_id)
+                    pbar.set_description(f'decoding modality [{curr_modality_id}]')
+
+                    latent_dim = self.dim_latents[curr_modality_id]
+                    noise = torch.randn((modality_length, latent_dim), device = device)
+
+                    def ode_step_fn(step_times, denoised):
+                        step_times = rearrange(step_times, ' -> 1 1') # batch size of 1
+                        step_times = F.pad(step_times, (num_past_modalities, 0), value = 1.) # past decoded modalities receive a time conditioning of 1.
+
+                        embeds = self.forward(
+                            [[*modality_sample, (curr_modality_id, denoised)]],
+                            times = step_times,
+                            return_embed = True,
+                        )
+
+                        to_flow_pred = self.model_to_latent_preds[curr_modality_id]
+                        flow = to_flow_pred(embeds)
+
+                        return flow[0, -modality_length:]
+
+                    times = torch.linspace(0, 1, modality_steps, device = device)
+                    trajectory = self.odeint_fn(ode_step_fn, noise, times)
+
+                    # add the sampled modality tokens
+
+                    sampled_modality = trajectory[-1]
+                    modality_sample.append((curr_modality_id, sampled_modality))
+
+                    # add the appropriate [eom]
+
+                    eom_id = self.eom_ids[curr_modality_id]
+                    modality_sample.append(tensor([eom_id], device = device))
+
+                    # back to decoding text
+
+                    pbar.update(modality_length)
+                    curr_length += modality_length
+
+                    num_past_modalities += 1
+                    curr_modality_id = None
+                    is_decoding_text = True
+
+        return modality_sample
+
     def forward(
         self,
-        modalities: list[list[Int['_'] | Float['_ _'] | tuple[int, Float['_ _']]]],
-        times: Float['b m'] | None = None,
+        modalities: list[ModalitySample],
+        times: (
+            Float['b m'] |
+            Callable[[Int['b m 3']], Float['b m']] | # allows a researcher to customize the times (noise level) based on the overall modality configuration of a sample
+            None
+        ) = None,
         return_loss = True,
-        return_breakdown = False
+        return_breakdown = False,
+        return_embed = False
     ) -> (
         Float['b n l'] |
+        Float['b n d'] |
         Float[''] |
         tuple[Float[''], LossBreakdown]
     ):
+        return_loss &= not return_embed
+
         device = self.device
 
+        # add "sentence" start and end tokens when training
+
+        if return_loss:
+            for modality in modalities:
+                modality.insert(0, tensor([self.sos_id], device = device))
+                modality.append(tensor([self.eos_id], device = device))
+
         # process list of text and modalities interspersed with one another
 
         modality_positions = []
@@ -622,6 +807,7 @@ class Transfusion(Module):
             batch_modality_positions = []
             batch_modality_tokens = []
             batch_text = []
+            offset = 0
 
             for modality in batch_modalities:
                 # if non-text modality detected and not given as a tuple
@@ -638,28 +824,36 @@ class Transfusion(Module):
                     modality_type, modality_tensor = modality
 
                     assert 0 <= modality_type < self.num_modalities, f'received a modality index that is out of range. only {self.num_modalities} modalities specified'
-                    assert self.dim_latents[modality_type] == modality_tensor.shape[-1], 'mismatch for modality latent dimension - expected {self.dim_latents[modality_type]} but received {modality_tensor.shape[-1]}'
+                    assert self.dim_latents[modality_type] == modality_tensor.shape[-1], f'mismatch for modality latent dimension - expected {self.dim_latents[modality_type]} but received {modality_tensor.shape[-1]}'
+
+                # auto move modality tensor to device of model
+
+                modality_tensor = modality_tensor.to(device)
 
-                offset = 0
                 length = modality_tensor.shape[0]
 
+                # handle text
+
                 if is_text:
                     batch_text.append(modality_tensor)
-                else:
+                    offset += length
+                    continue
 
-                    text_tensor = torch.full((length,), -1, device = device) # text is all -1 here, so text labels are not learned on
+                # otherwise handle a modality
 
-                    # add the [som] and [eom] tokens for the modality type
+                text_tensor = torch.full((length,), -1, device = device) # text is all -1 here, so text labels are not learned on
 
-                    som_id, eom_id = self.som_ids[modality_type], self.eom_ids[modality_type]
-                    text_tensor = F.pad(text_tensor, (1, 0), value = som_id)
-                    text_tensor = F.pad(text_tensor, (0, 1), value = eom_id)
+                # add the [som] and [eom] tokens for the modality type
 
-                    batch_text.append(text_tensor)
-                    batch_modality_tokens.append(modality_tensor)
-                    batch_modality_positions.append((modality_type, offset + 1, length)) # offset + 1 due to extra [som] token
+                som_id, eom_id = self.som_ids[modality_type], self.eom_ids[modality_type]
+                text_tensor = F.pad(text_tensor, (1, 0), value = som_id)
+                text_tensor = F.pad(text_tensor, (0, 1), value = eom_id)
 
-                offset += length
+                batch_text.append(text_tensor)
+                batch_modality_tokens.append(modality_tensor)
+                batch_modality_positions.append((modality_type, offset + 1, length)) # offset + 1 due to extra [som] token
+
+                offset += length + 2 # +2 due to [som] and [eom]
 
             text.append(torch.cat(batch_text))
             modality_tokens.append(batch_modality_tokens)
@@ -684,6 +878,11 @@ class Transfusion(Module):
         if modality_positions.shape[-1] == 2: # Int['b m 2'] -> Int['b m 3'] if type is not given (one modality)
             modality_positions = F.pad(modality_positions, (1, 0), value = 0)
 
+        # for now use dummy padding modality position info if empty (all zeros)
+
+        if modality_positions.numel() == 0:
+            modality_positions = F.pad(modality_positions, (0, 0, 0, 1))
+
         # embed the list of modality tokens into a sequence of Float['b n d'] at right offsets and lengths as dictated by modalities info tensor
 
         if torch.is_tensor(modality_tokens):
@@ -730,14 +929,17 @@ class Transfusion(Module):
         # noise the modality tokens
 
         if not exists(times):
-            times = torch.rand((batch, num_modalities), device = device)
+            if callable(times): # todo: rename to another field (derive_times: Callable?)
+                times = times(modality_positions)
+            else:
+                times = torch.rand((batch, num_modalities), device = device)
+
+        times = einsum(is_modalities.float(), times, 'b t m n, b m -> b t n')
 
         if return_loss:
             noised_modality_tokens = []
             flows = []
 
-            times = einsum(is_modalities.float(), times, 'b t m n, b m -> b t n')
-
             for modality_id, one_modality_tokens in enumerate(modality_tokens):
                 noise = torch.randn_like(one_modality_tokens)
 
@@ -783,6 +985,11 @@ class Transfusion(Module):
             modality_positions = modality_positions
         )
 
+        # early return for embedding for decoding modality
+
+        if return_embed:
+            return embed
+
         # text unembedding
 
         text_logits = self.to_text_logits(embed)
